function createRandomIdentifier(len) {
	let str = "";
	let bytes = [];

	if (typeof crypto === "object" && typeof crypto.getRandomValues === "function") {
		bytes = new Uint8Array(len);

		crypto.getRandomValues(bytes);
	} else {
		for (let i = 0; i < len; ++i) {
			// will never exceed 255
			bytes.push(Math.floor(Math.random() * 256));
		}
	}

	for (let i = 0; i < len; ++i) {
		str += (bytes[i]).toString(16).slice(0, 1);
	}

	// sanity check output
	if (str.length !== len) {
		throw new Error(`createRandomIdentifier(${len}) failed for unknown reasons.`)
	}

	return str
}

function createPromise() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

function createAsyncMutex() {
	let current_promise = null;

	return {
		async acquire() {
			if (current_promise !== null) {
				await current_promise;
			}

			let unlock;

			current_promise = new Promise(resolve => {
				unlock = resolve;
			});

			return () => {
				unlock();
				current_promise = null;
			}
		}
	}
}

function sendSingleShotRequestWithTimeout(instance, data, timeout = 0) {
	// create a request id that identifies this particular request
	const request_id = createRandomIdentifier(12);
	// create the promise returned by sendSingleShotRequest()
	const request_promise = createPromise();
	// keep track of timer to clear the timeout when response arrives
	let timeout_timer = null;

	if (timeout > 0) {
		timeout_timer = setTimeout(() => {
			const {reject} = request_promise;

			setTimeout(reject, 0, new Error(`Timeout of '${timeout}ms' reached.`));

			instance.open_requests.delete(request_id);
		}, timeout);
	}

	instance.open_requests.set(request_id, {
		request_promise,
		timer: timeout_timer
	});

	instance.sendJSONData({
		cmd: "request",
		request_id,
		data
	});

	return request_promise.promise
}

function sendRetransmittingRequest(instance, data, max_attempts = -1) {
	// create a request id that identifies this particular request
	const request_id = createRandomIdentifier(12);
	// create the promise returned by sendRequest()
	const request_promise = createPromise();
	// save retransmission delay for this request
	const retransmission_delay = instance.retransmission_delay;

	let attempt_no = 0;

	const retransmit = () => {
		if (!instance.open_requests.has(request_id)) {
			// no need for retransmit, request does not exist anymore
			// this happens when the request is either fullfilled or
			// close() was called.

			instance.debug(`no need for retransmitting request '${request_id}'`);

			return
		}

		const open_request = instance.open_requests.get(request_id);

		++attempt_no;

		if (attempt_no > max_attempts && max_attempts >= 0) {
			const error_msg = `Max retransmission attempts of ${max_attempts} reached.`;

			instance.debug(error_msg);

			setTimeout(request_promise.reject, 0, new Error(error_msg));

			instance.open_requests.delete(request_id);

			return
		}

		instance.debug(`retransmitting request '${request_id}'`);

		instance.sendJSONData({
			cmd: "request",
			request_id,
			data,
			max_attempts,
			attempt_no
		});

		open_request.timer = setTimeout(retransmit, retransmission_delay);
	};

	instance.open_requests.set(request_id, {
		request_promise,
		timer: setTimeout(retransmit, retransmission_delay)
	});

	instance.sendJSONData({
		cmd: "request",
		request_id,
		data,
		max_attempts,
		attempt_no
	});

	return request_promise.promise
}

function pruneHandledRequestsCache(instance) {
	const max_size = 1000;

	if (max_size > instance.handled_requests.size) {
		return
	}

	const map_keys = Array.from(instance.handled_requests.keys());
	const keys_to_be_removed = map_keys.slice(
		0, instance.handled_requests.size - max_size
	);

	for (const key of keys_to_be_removed) {
		instance.handled_requests.delete(key);
	}
}

async function onMessageReceived(instance, message) {
	//
	// call request handler
	//
	if (message.cmd === "request") {
		//
		// make sure every request is handled sequentially
		// this is needed for the response cache
		//
		const release = await instance.mutex.acquire();

		let response = null, from_cache = false;

		if (instance.handled_requests.has(message.request_id)) {
			response = instance.handled_requests.get(message.request_id);

			from_cache = true;
		} else {
			response = await instance.public_interface.requestHandler(message.data);
		}

		instance.sendJSONData({
			cmd: "response",
			original_request_id: message.request_id,
			response,
			from_cache
		});

		if (!from_cache) {
			instance.handled_requests.set(message.request_id, response);
		}

		pruneHandledRequestsCache(instance);

		await release();
	}
	//
	// handle incoming response
	//
	else if (message.cmd === "response") {
		const {original_request_id} = message;

		if (!instance.open_requests.has(message.original_request_id)) {
			instance.debug(`no pending request with id '${original_request_id}'`);

			return
		}

		const open_request = instance.open_requests.get(original_request_id);

		if (open_request.timer !== null) {
			clearTimeout(open_request.timer);
		}

		const {resolve} = open_request.request_promise;

		setTimeout(resolve, 0, message.response);

		instance.open_requests.delete(original_request_id);
	}
	else ;
}

function createRequestResponseProtocol(api, label = "") {
	const synchronize_token = createRandomIdentifier(32);
	let synchronized_promise = createPromise();

	let instance = {
		open_requests: new Map(),
		mutex: createAsyncMutex(),
		/*
		 * Keep a map of handled requests so if the same
		 * request is received a second time, the result is re-used
		 * and the request handler isn't called a second time.
		 */
		handled_requests: new Map(),
		ready: false,
		closed: false,

		retransmission_delay: 350,

		debug(...args) {
			if (instance.public_interface.debug !== true) return

			console.log(label, ...args);
		},

		sendJSONData(data) {
			api.sendMessage(`start{` + JSON.stringify(data) + `}end`);
		},

		messageHandler(message) {
			if (message.startsWith(`@anio-js-foundation/requestResponseProtocol:`)) {
				message = message.slice(`@anio-js-foundation/requestResponseProtocol:`.length);

				if (message.startsWith("sync:")) {
					const sync_token = message.slice("sync:".length);

					api.sendMessage(`@anio-js-foundation/requestResponseProtocol:ack:${sync_token}`);
				} else if (message.startsWith("ack:")) {
					const sync_token = message.slice("ack:".length);

					if (sync_token === synchronize_token && !instance.ready) {
						instance.ready = true;

						synchronized_promise.resolve();
					}
				} else {
					instance.debug("bad message", message);
				}
			} else if (instance.ready && message.startsWith("start{") && message.endsWith("}end")) {
				message = message.slice("start{".length);
				message = message.slice(0, message.length - "}end".length);

				onMessageReceived(instance, JSON.parse(message));
			} else {
				instance.debug("bad message", message);
			}
		},

		assertReadyAndNotClosed() {
			if (!instance.ready) {
				throw new Error(`Cannot send request when protocol is not ready.`)
			} else if (instance.closed) {
				throw new Error(`Cannot send request when protocol is closed.`)
			}
		},

		public_interface: {
			debug: false,

			ready() {
				return synchronized_promise.promise
			},

			setRetransmissionDelay(amount) {
				instance.retransmission_delay = amount;
			},

			requestHandler() {},

			withTimeout(timeout_value) {
				return {
					sendRequest(request_data) {
						let max_attempts = Math.floor(timeout_value / instance.retransmission_delay);

						instance.debug(
							`Calculated max_attempts = '${max_attempts}' from timeout value '${timeout_value}ms' (retransmission delay is '${instance.retransmission_delay}ms')`
						);

						return sendRetransmittingRequest(instance, request_data, max_attempts)
					},

					sendSingleShotRequest(request_data) {
						instance.assertReadyAndNotClosed();

						return sendSingleShotRequestWithTimeout(instance, request_data, timeout_value)
					}
				}
			},

			withMaxRetransmissionAttempts(max_attempts) {
				return {
					sendRequest(request_data) {
						return sendRetransmittingRequest(instance, request_data, max_attempts)
					}
				}
			},

			sendRequest(request_data) {
				return sendRetransmittingRequest(instance, request_data, -1)
			},

			sendSingleShotRequest(request_data) {
				instance.assertReadyAndNotClosed();

				return sendSingleShotRequestWithTimeout(instance, request_data, 0)
			},

			closeAllPendingRequests(reason = `closeAllPendingRequests() was called.`) {
				for (const [request_id, open_request] of instance.open_requests) {
					const {reject} = open_request.request_promise;

					setTimeout(reject, 0, new Error(reason));

					// make sure to cancel timer
					if (open_request.timer !== null) {
						clearTimeout(open_request.timer);
					}

					instance.open_requests.delete(request_id);
				}
			},

			close() {
				if (instance.closed) return

				instance.closed = true;

				setTimeout(
					instance.public_interface.closeAllPendingRequests, 0, "close() was called."
				);

				api.removeEventListener("message", instance.messageHandler);
			}
		}
	};

	api.on("message", instance.messageHandler);

	/**
	 * Slave can miss this message so repeat it
	 * every once in a while to get the master and slave
	 * synced up.
	 */
	const delay_map = [50, 50, 50, 100, 100, 150, 150, 150, 250, 500, 750, 1000];
	let delay_index = 0;

	const synchronize = async () => {
		if (instance.closed) return
		if (instance.ready) return

		instance.debug("attempt to synchronize");

		await api.sendMessage(`@anio-js-foundation/requestResponseProtocol:sync:${synchronize_token}`);

		// by default use the last value from the delay map
		let amount = delay_map[delay_map.length - 1];

		if (delay_map.length > delay_index) {
			amount = delay_map[delay_index];

			++delay_index;
		}

		instance.debug(`will be calling synchronize() in ${amount}ms`);

		setTimeout(synchronize, amount);
	};

	setTimeout(synchronize, 0);

	return instance.public_interface
}

export { createRequestResponseProtocol as default };
